# -*- coding: utf-8 -*-
# ====================================================================
# Copyright (c) 2021 Open Source Applications Foundation.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
# ====================================================================
#

import sys, os

from unittest import TestCase, main
from icu import *

class TestRuleBasedNumberFormat(TestCase):

    def testSpellout(self):

        formatter = RuleBasedNumberFormat(
            URBNFRuleSetTag.SPELLOUT, Locale("fr_FR"))

        text = formatter.format(80, "%spellout-ordinal-masculine")
        self.assertEqual(text, u'quatre-vingtième')
        text = formatter.format(76, '%spellout-ordinal-feminine-plural')
        self.assertEqual(text, u'soixante-seizièmes')
                
        formatter = RuleBasedNumberFormat(
            URBNFRuleSetTag.SPELLOUT, Locale("fr_CH"))

        text = formatter.format(70)
        self.assertEqual(text, u'septante')
        text = formatter.format(76, '%spellout-ordinal-feminine-plural')
        self.assertEqual(text, u'septante-sixièmes')
                

if __name__ == "__main__":
    main()
