# -*- coding: utf-8 -*-
# ====================================================================
# Copyright (c) 2023 Open Source Applications Foundation.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
# ====================================================================

import sys, os, six

from unittest import TestCase, main
from icu import *

class TestRegexMatcher(TestCase):

    def testReset(self):

        s = "Thuɔŋjäŋ"
        us = UnicodeString(s)
        matcher = RegexMatcher("t.+", URegexpFlag.CASE_INSENSITIVE)
        matcher.reset(us)
        self.assertTrue(matcher.matches())

        matcher.reset(s)
        self.assertTrue(matcher.matches())

        s = "དཔལ་དགོན་ཀླུ་གྲུབ།"
        us = UnicodeString(s)
        matcher = RegexMatcher('\\X')
        matcher.reset(us)
        results = []
        finds = ['ད', 'པ', 'ལ', '་', 'ད', 'གོ', 'ན', '་', 'ཀླུ', '་', 'གྲུ', 'བ', '།']
        while matcher.find():
            results.append(matcher.group())
        self.assertEqual(finds, results)

        matcher.reset(s)
        results = []
        while matcher.find():
            results.append(matcher.group())
        self.assertEqual(finds, results)

        
if __name__ == "__main__":
    main()
